#! /usr/bin/perl -w

$lang_ext = '';    # extension for language-specific-files (e.g. '-DE')
$show_names = 1;   # true/false; as displaying text in VRML eats CPU-power

# do what you want with this source, but don't blame me ;)
# for questions, comments, additions... mail to 'ninja@the-dreams.de'

use constant BRIGHTNESS => 0;
use constant FULLNAME => 0;
use constant POS_X => 1;
use constant POS_Y => 2;
use constant POS_Z => 3;
use constant AMOUNT => 4;

use constant PI => 4 * atan2(1,1);
use constant DEG2RAD => PI / 180;
use constant BRIGHT_MIN => 2;
use constant BRIGHT_MAX => 6;

sub acos { atan2(sqrt(1 - $_[0] * $_[0]), $_[0]) }
sub round {
    (my $ret = sprintf("%.3f", $_[0])) =~ s/\.*0*$//;
    $ret =~ s/0\././;
    return $ret;
}


open (CATALOG, './star_catalog');

$idx = 0;
while (<CATALOG>) {
    if (($bright = substr($_, 102, 5)) eq '     ' ) {
	$stars[$idx][BRIGHTNESS] = BRIGHT_MAX;
	$stars[$idx][POS_X] = 0;
	$stars[$idx][POS_Y] = 0;
	$stars[$idx][POS_Z] = 0;
    } else {
	if ($bright < BRIGHT_MIN) { $bright = BRIGHT_MIN; }
	if ($bright > BRIGHT_MAX) { $bright = BRIGHT_MAX; }
        $longitude =  substr($_,  90, 6) * DEG2RAD;
	$latitude  = (substr($_,  96, 6) + 90) * DEG2RAD;
	$stars[$idx][BRIGHTNESS] = $bright;
        $stars[$idx][POS_X] = cos($longitude) * sin($latitude);
        $stars[$idx][POS_Y] = sin($longitude) * sin($latitude);
        $stars[$idx][POS_Z] = cos($latitude);
    }
    $idx++;
}
close CATALOG;


open (VRML, ">./stars_n_constellations$lang_ext.wrl");
print VRML <<VRML_END;
#VRML V2.0 utf8

WorldInfo {
    title "Stars'n'Constellations in VRML97 V1.0 by Wolfram Sang"
    info [
	"star data from the 'Bright Star Catalog 5P'"
	"constellation data based on the program 'Cartes du Ciel'"
	"do what you want with this file but don't blame me for anything ;)"
    ]
}
Shape {
    geometry PointSet {
	coord DEF COORDS Coordinate {
	    point [
VRML_END

foreach (@stars) {
    $x = round($_->[POS_X]);
    $y = round($_->[POS_Y]);
    $z = round($_->[POS_Z]);
    print VRML "$x $y $z\n";
}
print VRML <<VRML_END;
	    ]
	}
	color Color {
	    color [
VRML_END
	    
$bright_conv = 7 / 10 / (BRIGHT_MAX - BRIGHT_MIN);
foreach (@stars) {
    $bright = round( 1 - ($_->[BRIGHTNESS] - BRIGHT_MIN) * $bright_conv);
    print VRML "$bright $bright $bright\n";
}
print VRML <<VRML_END;
	    ]
	}
    }
}
Shape {
    geometry IndexedLineSet {
	coord USE COORDS
	coordIndex [
VRML_END

open (CONSTELLATIONS, './constellations');
$name = '';
while (<CONSTELLATIONS>) {
    if (/^([A-Z]{3})/) {
	$star1 = substr($_,  6 ,4) - 1;
	$star2 = substr($_, 13 ,4) - 1;
	print VRML "$star1 $star2 -1\n";
	if ($1 ne $name) {
	    $name = $1;
	    $constellations{$name}[POS_X] = 0;
	    $constellations{$name}[POS_Y] = 0;
	    $constellations{$name}[POS_Z] = 0;
	    $constellations{$name}[AMOUNT] = 0;
	}
	$constellations{$name}[POS_X] += $stars[$star1][POS_X] + $stars[$star2][POS_X];
	$constellations{$name}[POS_Y] += $stars[$star1][POS_Y] + $stars[$star2][POS_Y];
	$constellations{$name}[POS_Z] += $stars[$star1][POS_Z] + $stars[$star2][POS_Z];
        $constellations{$name}[AMOUNT]++;
    }
}
close (CONSTELLATIONS);

print VRML <<VRML_END;
	]
	color Color {
	    color [
VRML_END

foreach (sort keys %constellations) {
    $blue = round(0.3 + rand(0.7));
    do {
	$intensity = round(rand(0.5));
    } until (abs($blue - $intensity) > 0.2);
    $counter = $constellations{$_}[AMOUNT];
    while ($counter--) {
        print VRML "$intensity $intensity $blue\n";
    }
}
print VRML <<VRML_END;
	    ]
	}
    colorPerVertex FALSE
    }
}
VRML_END


open (CONST_NAMES, "./constellation_names$lang_ext");
while (<CONST_NAMES>) {
    if (/^([A-Z]{3}) ([^\n\r]*)/) {
	$constellations{$1}[FULLNAME] = $2;
    }
}
close CONST_NAMES;    

if ($show_names) {
    print VRML <<VRML_END;
PROTO Name [
    field MFString name []
]
{
    Shape {
        geometry Text {
    	string IS name
	    fontStyle FontStyle {
		size 0.05
		family "SANS"
	    }
	}
    }
}    
VRML_END
    foreach (sort keys %constellations) {
	if ($constellations{$_}[AMOUNT]) {
	    $x = $constellations{$_}[POS_X];
	    $y = $constellations{$_}[POS_Y];
	    $z = $constellations{$_}[POS_Z];
	    $xyz_length = sqrt($x * $x + $y * $y + $z * $z);
	    $x = round($x / $xyz_length);
	    $y = round($y / $xyz_length);
	    $z = round($z / $xyz_length);
	    $constellations{$_}[FULLNAME] =~ /(.*?)(?: \(|\z)/;
	    print VRML <<VRML_END;
Transform {
    translation $x $y $z
    children Billboard {
	axisOfRotation 0 0 0
	children Name {
	    name "$1"
	}
    }
}
VRML_END
	}
    }
}    

foreach (sort {$constellations{$a}[FULLNAME] cmp $constellations{$b}[FULLNAME]} 
 keys %constellations) {
    if ($amount = $constellations{$_}[AMOUNT]) {
	$z = $constellations{$_}[POS_Z] / ($amount * 2);
	$norm_x = $constellations{$_}[POS_Y];
	$norm_y = -$constellations{$_}[POS_X];
        $norm_z = 0;
        $norm_length = sqrt($norm_x * $norm_x + $norm_y * $norm_y);
        ($norm_x = sprintf ("%.6f", $norm_x / $norm_length)) =~ s/0\././;
        ($norm_y = sprintf ("%.6f", $norm_y / $norm_length)) =~ s/0\././;
        $norm_deg = round(acos(-$z));
        print VRML <<VRML_END;
Viewpoint {
    position 0 0 0
    orientation $norm_x $norm_y $norm_z $norm_deg
    description "$constellations{$_}[FULLNAME] [$_]"
}
VRML_END
    } #else { print "Missing data for '$constellations{$_}[FULLNAME]'!\n"; }
}

close VRML;
